/*
    Copyright 2010 Luis Miguel Merino Bonilla

    This file is part of GNU Taste

    GNU Taste is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    GNU Taste is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with GNU Taste. If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef 
__DATABASE__
#define __DATABASE__

#include "hash.h"
#include "feature_vector.h"
#include "similarity.h"
#include "storage.h"
#include "log.h"
#include <cstring>
#include <fstream>

class Database{
private:
	Storage *disk;
	bool normalize;
public:
	string name;
	Hash<feature_vector> users;
	Hash<feature_vector> items;
	Hash<float> userSum;
	Hash<int> userCount;
	int numRatings;
	int lastUpdate;

	Database(string n) : name(n), numRatings(0), lastUpdate(0), normalize(false) {
		users = Hash<feature_vector>(100001);
		items = Hash<feature_vector>(100001);
		disk = new Storage();
	}

	void insertRating(int user, int item, float value, int tstamp){
		//users[user].insert(item, value);
		//items[item].insert(user, value);
		getUser(user)->insert(item, value);
		getItem(item)->insert(user, value);
		userSum[user] += value;
		++userCount[user];
	}

	void loadStats(){
		disk->loadHash("data/sum.hash", userSum);
		disk->loadHash("data/count.hash", userCount);
	}

	void saveToStore(){
		log("Almacenando perfiles de usuario...");
		Hash<feature_vector>::Iterator it = users.begin();
		while(!it.end()){
			it->value.sort();
			disk->saveProfile(it->key, it->value);
			++it;
		}
		log("Almacenando items...");
		Hash<feature_vector>::Iterator it2 = items.begin();
		while(!it2.end()){
			it2->value.sort();
			disk->saveItem(it2->key, it2->value);
			++it2;
		}

		log("Almacenando tabla de contadores...");
		disk->saveHash("data/sum.hash", userSum);
		log("Almacenando tabla de acumuladores...");
		disk->saveHash("data/count.hash",userCount);
		
	}

	feature_vector * getUser(int u){
		//Intentamos obtener el vector de memoria
		feature_vector * ob = users.get(u);

		//Si no está cargado, lo traemos desde disco
		if (!ob){
			//cout << "El usuario #"<<u<<" no está en memoria. Cargando..." << endl;
			ob = disk->loadProfile(u);
			users.insert(u,*ob);
		}
		//cout << *ob;
		return ob;
	}


	feature_vector * getItem(int i){
		//Intentamos obtener el vector de memoria
		feature_vector * ob = items.get(i);

		//Si no está cargado, lo traemos desde disco
		if (!ob){
			//cout << "El item #"<<i<<" no está en memoria. Cargando..." << endl;
			ob = disk->loadItem(i);
			if(this->normalize)
				this->normalizeItem(i, ob);
			items.insert(i,*ob);
		}

		return ob;
	}

	void setNormalizing(bool state){
		this->normalize = state;
	}

	void normalizeItem(int id, feature_vector *f){
		for(node *n = f->begin(); n != 0; n = n->next){
			n->value -= this->getMean(n->id);
		}
	}

	float getMean(int u){
		return userSum[u]/userCount[u];
	}

	void show(){
		cout << users;
	}
	
};

class Loader{
public:
	virtual void load(char *fname) = 0;
};



class MovielensLoader : public Loader{
private:
	Database * db;

public:
	MovielensLoader(Database *d) : db(d) {
		log("Inicializando MovielensLoader...");
	}

	void load(char *fname){
		FILE *f_test = fopen(fname, "r");
		int user, item, tstamp;
		float value;
		log(string("Cargando fichero de ratings ") + string(fname));

//		while (fscanf(f_test, "%d::%d::%f::%d",&user,&item,&value,&tstamp) != EOF){
		while (fscanf(f_test, "%d %d %f %d",&user,&item,&value,&tstamp) != EOF){
			//cout << "Cargando " << user <<":"<<item<<":"<<value<<":"<<tstamp<<endl;
			db->insertRating(user, item, value, tstamp);
		}
	}
};

#endif
